import 'package:flutter/widgets.dart';
import 'package:flutter_test/flutter_test.dart';
import 'package:fst/fst.dart';
import 'package:fst_test/fst_test.dart';

void main() {
  group('ProviderLeaf', () {
    testWidgets('provides a value via BuildContext', (tester) async {
      await tester.pumpStateTree(
        (context, t) {
          t.provide(42);
        },
        StateTreeBuilder(builder: (context, t, _) {
          final value = t.consume<int>();
          return Container(key: ValueKey(value));
        }),
      );

      expect(find.byKey(const ValueKey(42)), findsOneWidget);
    });

    testWidgets('notifies consumers when value changes', (tester) async {
      final setValue = await tester.pumpStateTree(
        (context, t) {
          final (v, setV) = t.state(() => 42);
          t.provide(v);

          return setV;
        },
        StateTreeBuilder(builder: (context, t, _) {
          final value = t.consume<int>();
          return Container(key: ValueKey(value));
        }),
      );

      expect(find.byKey(const ValueKey(42)), findsOneWidget);

      setValue(43);

      await tester.pumpStateTree();
      expect(find.byKey(const ValueKey(43)), findsOneWidget);
    });

    testWidgets(
      'throws when same type values are provided without a token',
      (tester) async {
        await tester.pumpStateTree(
          (context, t) {
            t.provide(42);
            t.provide(43);
          },
          StateTreeBuilder(builder: (context, t, _) {
            t.consume<int>();
            t.consume<int>();
            return Container();
          }),
        );

        expect(tester.takeException(), isA<StateError>());
      },
    );

    testWidgets(
      'throws if no providers in tree, but value is consumed',
      (tester) async {
        await tester.pumpStateTree((context, t) {
          t.consume<int>();
        });

        expect(tester.takeException(), isA<StateError>());
      },
    );

    testWidgets(
      'throws if there is no value of requested type provided',
      (tester) async {
        final setValue = await tester.pumpStateTree(
          (context, t) {
            final (v, setV) = t.state<int?>(() => null);
            if (v != null) {
              t.provide<int>(v);
            }

            return setV;
          },
          StateTreeBuilder(builder: (context, t, _) {
            t.consume<String>();
            return Container();
          }),
        );

        expect(tester.takeException(), isA<StateError>());

        setValue(42);

        await tester.pumpStateTree();
        expect(tester.takeException(), isA<StateError>());
      },
    );

    testWidgets('allows to query provided value', (tester) async {
      final setState = await tester.pumpStateTree(
        (context, t) {
          final (value, setState) = t.state(() => 0);
          t.provide(value);

          return setState;
        },
        StateTreeBuilder(builder: (context, t, _) {
          final value = t.query<bool, int>((v) => v.isEven);
          return Container(key: ValueKey(value));
        }),
      );

      expect(find.byKey(const ValueKey(true)), findsOneWidget);

      setState(1);
      await tester.pumpStateTree();
      expect(find.byKey(const ValueKey(true)), findsNothing);

      setState(2);
      await tester.pumpStateTree();
      expect(find.byKey(const ValueKey(true)), findsOneWidget);
    });

    testWidgets(
      'allows same type values to be provided and distinguished by token',
      (tester) async {
        await tester.pumpStateTree(
          (context, t) {
            t.provide(1, token: #a);
          },
          StateTreeBuilder(builder: (context, t, child) {
            t.provide(2, token: #b);

            return child!;
          }, child: StateTreeBuilder(builder: (context, t, _) {
            final valueA = t.consume<int>(token: #a);
            final valueB = t.consume<int>(token: #b);

            return Container(key: ValueKey(valueA + valueB));
          })),
        );

        expect(find.byKey(const ValueKey(3)), findsOneWidget);
      },
    );

    testWidgets('throws if no value was shadowed', (tester) async {
      await tester.pumpStateTree(
        (context, t) {
          t.provide(1, token: #a);
        },
        StateTreeBuilder(builder: (context, t, child) {
          t.provide(2, token: #b);

          return child!;
        }, child: StateTreeBuilder(builder: (context, t, _) {
          t.consume<int>();

          return Container();
        })),
      );

      expect(tester.takeException(), isA<StateError>());
    });

    testWidgets(
      'throws if no value provided with a given token',
      (tester) async {
        await tester.pumpStateTree(
          (context, t) {
            t.provide(1, token: #a);
          },
          StateTreeBuilder(builder: (context, t, _) {
            final v = t.consume<int>(token: #b);
            return Container(key: ValueKey(v));
          }),
        );

        expect(tester.takeException(), isA<StateError>());
      },
    );
  });
}
