import 'dart:async';

import 'package:flutter_test/flutter_test.dart';
import 'package:fst/fst.dart';
import 'package:fst_test/fst_test.dart';

typedef R = (Result<int> value, void Function(int) setDep);

void main() {
  group('StreamLeaf', () {
    testWidgets('binds stream', (tester) async {
      final ctrl = StreamController<int>();
      var value = await tester.pumpStateTree((context, t) {
        final v = t.stream(() => ctrl.stream);

        return v;
      });

      expect(value, isA<Pending<int>>());

      ctrl.add(42);
      value = await tester.pumpStateTreeAndSettle();
      expect(value, isA<Success<int>>());
      expect((value as Success<int>).value, 42);

      ctrl.add(43);
      value = await tester.pumpStateTreeAndSettle();
      expect(value, isA<Success<int>>());
      expect(value.value, 43);
    });

    testWidgets('updates stream when dependency is changed', (tester) async {
      final ctrl1 = StreamController<int>();
      final ctrl2 = StreamController<int>();

      var (value, setDep) = await tester.pumpStateTree<R>(
        (context, t) {
          final (d, setD) = t.state(() => 0);
          final v = t.stream(() {
            return switch (d) {
              0 => ctrl1.stream,
              1 => ctrl2.stream,
              _ => const Stream<int>.empty()
            };
          }, (d,));

          return (v, setD);
        },
      );
      expect(value, isA<Pending<int>>());

      ctrl1.add(42);
      (value, setDep) = await tester.pumpStateTreeAndSettle<R>();
      expect(value, isA<Success<int>>());
      expect((value as Success<int>).value, 42);

      setDep(1);
      (value, setDep) = await tester.pumpStateTreeAndSettle<R>();
      expect(value, isA<Pending<int>>());
      expect(ctrl1.hasListener, false);

      ctrl2.add(43);
      (value, setDep) = await tester.pumpStateTreeAndSettle<R>();
      expect(value, isA<Success<int>>());
      expect((value as Success<int>).value, 43);
    });

    testWidgets(
      'cancels stream subscription when no longer bound',
      (tester) async {
        final ctrl = StreamController<int>();

        var (value, cancel) =
            await tester.pumpStateTree<(Result<int>?, void Function())>(
          (context, t) {
            final (isBound, setIsBound) = t.state(() => true);
            void cancel() => setIsBound(false);

            if (isBound) {
              return (t.stream(() => ctrl.stream), cancel);
            }

            return (const Success(0), cancel);
          },
        );

        expect(value, isA<Pending<int>>());

        ctrl.add(42);
        (value, cancel) = await tester.pumpStateTreeAndSettle();
        expect(value, isA<Success<int>>());
        expect((value as Success<int>).value, 42);

        cancel();
        await tester.pumpStateTree();

        ctrl.add(43);
        await tester.pump();
        expect(tester.binding.hasScheduledFrame, false);
        expect(ctrl.hasListener, false);
      },
    );
  });
}
